/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.control;

import de.unibi.airobots.bonsai.skills.SensorActuatorManager;
import de.unibi.airobots.bonsai.skills.SkillStateMachine;
import de.unibi.airobots.bonsai.skills.view.StateMachineViewer;
import de.unibi.airobots.bonsai.skills.view.StateMachineViewerListener;
import java.util.Iterator;
import java.util.List;
import org.apache.commons.scxml.SCXMLListener;
import org.apache.commons.scxml.model.Transition;
import org.apache.commons.scxml.model.TransitionTarget;

/**
 * Controller for the SCXML state machine.
 *
 * @author lkettenb
 */
public class StateMachineController implements StateMachineViewerListener, 
        SCXMLListener {
    
    private String pathToConfig = null;
    private String pathToTask = null;
    
    private StateMachineViewer stateMachineViewer;
    private SensorActuatorManager sensorActuatorManager;
    private SkillStateMachine skillStateMachine;
    
    /**
     * Convenience constructor.
     */
    public StateMachineController() {
        
    }
    
    
    /**
     * Use this constructor if the path to task and configuration file is
     * already known.
     * 
     * @param pathToConfig Path to configuration file.
     * @param pathToTask Path to task file.
     */
    public StateMachineController(String pathToConfig, String pathToTask) {
        this.pathToConfig = pathToConfig;
        this.pathToTask = pathToTask;
    }
    
    
    /**
     * Initialize controller, setup GUI, add listener, etc.
     */
    public void initialize() {
        if (pathToConfig == null || pathToTask == null) {
            stateMachineViewer = new StateMachineViewer();
        } else {
            stateMachineViewer = 
                    new StateMachineViewer(pathToConfig, pathToTask);
        }
        stateMachineViewer.addListener(this);
    }
    
    
    @Override
    public void loadConfig(String pathToConfig) {
        sensorActuatorManager = new SensorActuatorManager();
        boolean success = sensorActuatorManager.initialize(pathToConfig);
        stateMachineViewer.setLoadConfigSuccessful(success);
        if (success) {
            String[] actuators = sensorActuatorManager.getActuators();
            String[] sensors = sensorActuatorManager.getSensors();
            stateMachineViewer.updateSensorsActuatorsListModel(sensors, actuators);
        }
    }

    
    @Override
    public void loadTask(String pathToTask) {
        skillStateMachine = new SkillStateMachine(sensorActuatorManager);
        boolean successful = skillStateMachine.initalize(pathToTask);
        stateMachineViewer.setLoadTaskSuccessful(successful);
        if (successful) {
            skillStateMachine.addListener(this);
        }
    }

    
    @Override
    public void executeStateMachine() {
        skillStateMachine.startMachine();
    }

    @Override
    public void onEntry(TransitionTarget tt) {
        List<Transition> a =  tt.getTransitionsList();
        String[] eventList = new String[a.size()];
        String[] nextStates = new String[a.size()];
        Iterator<Transition> iterator = a.iterator();
        for (int i = 0; iterator.hasNext(); i++) {
            Transition next = iterator.next();
            eventList[i] = next.getEvent();
            nextStates[i] = next.getNext();
        } 
        stateMachineViewer.updateEventList(eventList);
        stateMachineViewer.updateNextStatesList(nextStates);
        stateMachineViewer.addCurrentActiveState(tt.getId());
    }

    @Override
    public void onExit(TransitionTarget tt) {
        stateMachineViewer.removeCurrentActiveState(tt.getId());
    }

    @Override
    public void onTransition(TransitionTarget tt, TransitionTarget tt1, 
            Transition trnstn) {
        
    }

    @Override
    public void fireEvent(String event) {
        skillStateMachine.fireEvent(event, null);
    }
    
}
