/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.states.dialog.followme;

import de.unibi.airobots.bonsai.actuators.SpeechActuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.skills.SkillStateMachine;
import de.unibi.airobots.bonsai.skills.abstractstates.AbstractState;
import de.unibi.airobots.bonsai.skills.helper.SpeechManager;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.EndException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.ExecuteException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.InitException;
import de.unibi.airobots.btl.data.Utterance;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import org.apache.log4j.Logger;


/**
 * Use this class to understand the different "FolloMe" commands.
 * 
 * <p>
 * The below exit statuses will be fired: <code>successful.followMe</code>
 * if the robot shall follow a person, <code>successful.waitHere</code> if
 * the robot shall recognize a person and <code>successful</code> if 
 * nothing else is to do.
 * </p>
 * 
 * @author lkettenb
 */
public class WaitForCommand extends AbstractState {
    /*
     * What the robot will say.
     */
    private static final String FOLLOWING = "Ok, I will follow you";
    private static final String WAITING_TEN_SECONDS = 
            "Ok, I will wait here for ten seconds!";
    private static final String WAITING = 
            "Ok, I will wait here until you are back!";
    private static final String FORCE_FOLLOWING = 
            "Do you want me to follow you?";
    private static final String CONFIRM_END = "Do you want me to shut down?";
    private static final String END = "Good bye, have a nice day!";
    
    /*
     * Actuators used by this state.
     */
    private SpeechActuator speechActuator;
    
    /*
     * Sensors used by this state.
     */
    private Sensor<Utterance> speechSensor;
    
    /**
     * The logger.
     */
    private Logger logger = super.getLogger();
    
    /**
     * SpeechManager used for understanding etc.
     */
    private SpeechManager speechManager;
    
    /**
     * Number of possible misunderstandings.
     */
    private static final int MAX_MISUNDERSTANDINGS = 15;
    
    /**
     * Time to wait for command.
     */
    private static final int MAX_WAIT_TIME = 60000;
    
    /**
     * After this time interaction is forced.
     */
    private static final int WAIT_PHRASE_WAIT_TIME = 1000000;
    
    /**
     * Time to wait (usually 10 seconds) when robot is told to wait.
     */
    private static final int WAIT_TIME = 10000;

    private List<String> waitingPhrases = Arrays.asList(
            "Please tell me what to do!", "What do you want me to do?",
            "What can i do for you?");

    private int nothingUnderstoodCounter = 0;
    private long lastWaitingPhrase = 0;
    private long conversationStartTime = System.currentTimeMillis();
    
    /**
     * Constructor.
     * @param aStateMachine The state machine to use.
     */
    public WaitForCommand(SkillStateMachine aStateMachine) {
        super(aStateMachine);
    }

    
    @Override
    public ExitStatus init() throws InitException {
        speechSensor = getManagedSensor("SpeechSensor");
        speechActuator = getManagedActuator("SpeechActuator");
        speechManager = new SpeechManager(speechSensor, speechActuator);
        logger.info("Waiting for commands ...");
        return ExitStatus.SUCCESS;
    }

    
    @Override
    public ExitStatus execute() throws ExecuteException {
        try {
            // force interaction
            if ((System.currentTimeMillis() - lastWaitingPhrase) > 
                    WAIT_PHRASE_WAIT_TIME) {
                Collections.shuffle(waitingPhrases);
                say(waitingPhrases.get(0));
                lastWaitingPhrase = System.currentTimeMillis();
            }

            // wait for commands
            boolean understoodSomething = 
                    speechManager.waitForUnderstanding(false, 100, 3000);
            if (!understoodSomething) {
                setProcessingStatus(getWaitProcessingMessage());
                return ExitStatus.SUCCESS;
            }
            
            Set<String> understood = speechManager.getNonTerminals();
            logger.info(understood);
            if (understood.contains("followMe")) {
                nothingUnderstoodCounter = 0;
                conversationStartTime = System.currentTimeMillis();
                logger.info("Understood followMe");
                say(FOLLOWING);
                setProcessingStatus("followMe");
                return ExitStatus.SUCCESS;
            } else if (understood.contains("waitTenSeconds")) {
                nothingUnderstoodCounter = 0;
                conversationStartTime = System.currentTimeMillis();
                logger.info("Understood waitTenSeconds");
                say(WAITING_TEN_SECONDS);
                try {
                    sleep(WAIT_TIME);
                } catch (InterruptedException e) {
                    logger.warn("ignoring interuption");
                }
                setProcessingStatus("followMe");
                return ExitStatus.SUCCESS;
            } else if (understood.contains("waitHere")) {
                nothingUnderstoodCounter = 0;
                conversationStartTime = System.currentTimeMillis();
                logger.info("Understood waitHere");
                say(WAITING);
                setProcessingStatus("waitHere");
                return ExitStatus.SUCCESS;
            } else if (understood.contains("bye")) {
                nothingUnderstoodCounter = 0;
                conversationStartTime = System.currentTimeMillis();
                logger.info("Understood bye");
                say(CONFIRM_END);

                speechManager.waitForUnderstanding(false);
                understood = speechManager.getNonTerminals();
                if (understood.contains("confirm_yes")) {
                    say(END);
                }
                return ExitStatus.SUCCESS;
            } else {
                logger.debug("Understood nothing");
                nothingUnderstoodCounter++;
            }

            // force following after too many misunderstandings
            long timeDiff = System.currentTimeMillis() - conversationStartTime;
            if (nothingUnderstoodCounter >= MAX_MISUNDERSTANDINGS
                    || timeDiff > MAX_WAIT_TIME) {

                say(FORCE_FOLLOWING);
                nothingUnderstoodCounter = 0;
                conversationStartTime = System.currentTimeMillis();

                speechManager.waitForUnderstanding(false);
                understood = speechManager.getNonTerminals();

                if (understood.contains("confirm_yes")
                        || understood.contains("followMe")) {
                    logger.debug("Understood confirm_yes");
                    say(FOLLOWING);
                    setProcessingStatus("followMe");
                    return ExitStatus.SUCCESS;
                }
            }
            setProcessingStatus(getWaitProcessingMessage());
            return ExitStatus.SUCCESS;
        } catch (IOException ex) {
            // To bad. Unable to understand anything.
            throw new ExecuteException(ex.getMessage());
        }
    }

    
    @Override
    public ExitStatus end() throws EndException {
        return ExitStatus.SUCCESS;
    }
    
    /**
     * Use speech actuator to say something and catch IO exception.
     * 
     * @param text Text to be said.
     */
    private void say(String text) {
        try {
            speechActuator.say(text);
        } catch (IOException ex) {
            // Not so bad. The robot just says nothing.
            logger.warn(ex.getMessage());
        }
    }
}
