/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.strategies.follow;

import java.io.IOException;
import java.util.LinkedList;

import net.sf.xcf.transport.XOPData;
import nu.xom.Builder;
import nu.xom.ParsingException;
import nu.xom.ValidityException;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.actuators.TaskPatternActuator;
import de.unibi.airobots.bonsai.core.BonsaiManager;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.sensors.data.PersonList;
import de.unibi.airobots.bonsai.util.CoordinateSystemConverter;
import de.unibi.airobots.bonsai.util.helper.GoalReachedWaiter;
import de.unibi.airobots.bonsai.util.helper.GoalReachedWaiter.ReachStatus;
import de.unibi.airobots.bonsai.xcf.actuators.XcfTaskPatternActuator;
import de.unibi.airobots.btl.data.GlobalPlannerStateChange;
import de.unibi.airobots.btl.data.NavigationGoalData;
import de.unibi.airobots.btl.data.PersonData;
import de.unibi.airobots.btl.data.PositionData;
import de.unibi.airobots.btl.data.SpeedData;
import de.unibi.airobots.btl.data.taskpattern.PleaseWaitNotification;
import de.unibi.airobots.btl.data.taskpattern.RobotLostHumanNotification;

/** 
 * This class refactors the old simple following strategy with adaption
 * to new demands.
 *
 * @author Denis Schulze <dschulze@techfak.uni-bielefeld.de>
 * @author vrichter, bcarlmey
 */
public class TaskPatternBasedNewPersonTooFarFollowingStrategy implements FollowStrategy {

	private final Logger logger = Logger.getLogger(getClass());	
    private Builder builder = new Builder();


	/** Some definitions for criterions. */
	private static final double MAX_ANGLE_DIFFERENCE_TO_CENTER = Math.PI / 4;

	private static final double MIN_DISTANCE_UNTIL_MOVE = 1000.;

	/** Tolerance values for navigation goals. */
	private static final double COORDINATE_TOLERANCE = 0.5;
	private static final double YAW_TOLERANCE = Math.PI * 0.3;
	
	/** The distance which the robot drives when stepForward is called. */
	private static final double STEPSIZE = 1;
	
	/** A history of navigations goals i.e. the position of a person and 
	 * the corresponding maximal stack size. */
	private final LinkedList<NavigationGoalData> lastKnownPositions = new LinkedList<NavigationGoalData>();
	private final int STACK_SIZE = 10;

	/** Sensor interfaces. */
	private final Sensor<GlobalPlannerStateChange> gplanChangeSensor;
	private final Sensor<PersonList> personSensor;
	private final Sensor<PositionData> posSensor;
	private final Sensor<SpeedData> speedSensor;

	/** Actuator interfaces. */
	private final NavigationActuator navActuator;
    private final TaskPatternActuator taskActuator;

	/** The id of the person to follow. */
	private int personId = -1;
	
	private int loop = 0;

																			
	/** Default constructor. */
	public TaskPatternBasedNewPersonTooFarFollowingStrategy() {
		personSensor			= BonsaiManager.getInstance().createSensor("PersonSensor", PersonList.class);
		posSensor					= BonsaiManager.getInstance().createSensor("PositionSensor", PositionData.class);
		navActuator				=	BonsaiManager.getInstance().createActuator("NavigationActuator", NavigationActuator.class);
		speedSensor				= BonsaiManager.getInstance().createSensor("SpeedSensor", SpeedData.class);
		taskActuator		= BonsaiManager.getInstance().createActuator("TaskPatternActuator", XcfTaskPatternActuator.class);
		gplanChangeSensor	= BonsaiManager.getInstance().createSensor("GlobalPlannerStateChangeSensor",GlobalPlannerStateChange.class);	
	}	


	@Override
	public void init() {		
	}

	@Override
	public void execute() throws IOException {
			
		try {

			// If we have no id of a person to follow, search a new one
			// otherwise find the old person to follow.	
			PersonData data = (personId == -1) ? searchNewPerson() : searchKnownPerson();
			if (data != null){
			    followPerson(data);   
			} else {
			    stepForward();
			}

		} catch (InterruptedException e) {
			logger.fatal("InterruptedExeption caugth: \n" + e.getMessage());
		}

	}

	@Override
	public void execute(int personId) throws IOException {

		try {

			this.personId = personId;
			PersonData data = searchKnownPerson();
            if (data != null){
                followPerson(data);   
            } else {
                stepForward();
            }
	
		} catch (InterruptedException e) {
			logger.fatal("InterruptedException caught: \n" + e.getMessage());
		} 

	}

    /** Searches for a new person to follow using the angle of persons to the center
	 * of the robot. The person with the minimal distance to the center of the robot
	 * is returned or null if no person matches the criterions. */
	private PersonData searchNewPerson() throws IOException, InterruptedException {

		logger.debug("Searching a new person to follow.");

		PersonList personList = personSensor.readLast();

		PersonData result = null;

		// Search nearest person to the center of the robot.
		double minAngle = MAX_ANGLE_DIFFERENCE_TO_CENTER;
		for (PersonData pData : personList) {

			double currentAngle = Math.abs(pData.getAngle());
			int id = pData.getId();
			if (currentAngle <= minAngle && id >= 0) {
				minAngle = currentAngle;
				result = pData;
			} 

		}

		// Check if a person was found.
		if (result == null && loop%50==0) {
		    try {
                taskActuator.initiate(new XOPData(builder.build("<"
                        + RobotLostHumanNotification.getBaseTag() + "/>",
                        null)));
            } catch (ValidityException e) {
                logger.error(e);
            } catch (ParsingException e) {
                logger.error(e);
            }
//			speechActuator.say("I think I have lost you!", true);
		}		
		loop++;
		return result;		

	}

	/** Tries to refind the person with the id personId. */
	private PersonData searchKnownPerson() throws IOException, InterruptedException {

		logger.debug("Searching known person to follow.");

		PersonList personList = personSensor.readLast();

		// Check if the old person is present in the current person list.
		PersonData pData = personList.getPersonById(personId);

		// If the person is not present in the list, drive to the last known
		// position and search a new one hoping that it is the old one.
		if (pData == null) {
		    try {
                taskActuator
                .initiate(new XOPData(builder.build("<"
                        + PleaseWaitNotification.getBaseTag()
                        + "/>", null)));
            } catch (ValidityException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            } catch (ParsingException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
//			speechActuator.say("Oh, please wait.");
			driveToLastKnownPosition();
			pData = searchNewPerson();
		}	

		return pData;

	}	

	/** Makes the robot drive to the last known position of a person. */
	private void driveToLastKnownPosition() throws IOException, InterruptedException {

		PositionData currentPosition = posSensor.readLast();

		// Find the last known position and drive to that coordinates.
		// Is this search really necessary???
		double minDist = Double.MAX_VALUE;
		NavigationGoalData nearestData = null;
		for (int i = 0; i < lastKnownPositions.size(); i++) {		
			NavigationGoalData data = lastKnownPositions.get(i);			
			double xDiff = currentPosition.getX() - data.getX();
			double yDiff = currentPosition.getY() - data.getY();
			double dist = Math.hypot(xDiff, yDiff);			
			if (dist < minDist) {
				minDist = dist;
				nearestData = data;
			}
		}

		if (nearestData == null) {
			logger.debug("No last known position available.");
			return;
		}

		// Drive to last known position.
		logger.debug("Driving to last known position.");

		nearestData.setCoordinateTolerance(COORDINATE_TOLERANCE);
		nearestData.setYawTolerance(YAW_TOLERANCE);
	
		navActuator.setGoal(nearestData);		

		GoalReachedWaiter waiter = new GoalReachedWaiter(
			posSensor, speedSensor, gplanChangeSensor, navActuator, 800, GoalReachedWaiter.DEFAULT_STEPS);
		ReachStatus goalReached = waiter.waitForReachingGoal(nearestData, 10000);
		if (goalReached == ReachStatus.REACHED) {
			logger.debug("last known position reached");
		} else {
			logger.debug("Could not reach position: " + goalReached);
		}

	}			

	/** Follow the assigned person. */
	private void followPerson(PersonData personData) throws IOException, InterruptedException {

		logger.debug("Following person.");

		double angle = personData.getAngle();
		double distance = personData.getDistance();

		PositionData currentPosition = posSensor.readLast();	

		// If the distance of the person is great enough, add a new navigation gaol.
		if (distance > MIN_DISTANCE_UNTIL_MOVE) {

			NavigationGoalData personGoal = CoordinateSystemConverter
						.polar2NavigationGoalData(currentPosition, angle,
								(distance - MIN_DISTANCE_UNTIL_MOVE) / 1000.0);

			if (lastKnownPositions.size() >= STACK_SIZE) {
				lastKnownPositions.poll();
			}
			lastKnownPositions.add(personGoal);

			logger.debug("Moving: " + (distance - MIN_DISTANCE_UNTIL_MOVE) + " m with angle: " + angle + " (Radiant)");

			personGoal.setCoordinateTolerance(COORDINATE_TOLERANCE);
			personGoal.setYawTolerance(YAW_TOLERANCE);
			navActuator.setGoal(personGoal);


		} else if (distance < MIN_DISTANCE_UNTIL_MOVE) {

				NavigationGoalData personGoal = CoordinateSystemConverter
						.polar2NavigationGoalData(currentPosition, angle, 0.0);

				logger.debug("Turning: " + angle + " (Radiant)");
				personGoal.setCoordinateTolerance(COORDINATE_TOLERANCE);
				personGoal.setYawTolerance(YAW_TOLERANCE);

				navActuator.setGoal(personGoal);

		}

	}
	
	/** Makes the robot move STEPSIZE forward. 
	 * @throws InterruptedException 
	 * @throws IOException 
	 * */
    private void stepForward() throws IOException, InterruptedException {
        logger.info("no person. making a step forward.");
        PositionData currentPosition = posSensor.readLast();    
        
        NavigationGoalData goal = CoordinateSystemConverter
        .polar2NavigationGoalData(currentPosition, 0.0, STEPSIZE);

        goal.setCoordinateTolerance(COORDINATE_TOLERANCE);
        goal.setYawTolerance(YAW_TOLERANCE);

        navActuator.setGoal(goal);
    }


}

