/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.util;

/**
 * A combination of two objects of the same type. This pair is immutable, but
 * the contained objects are not, in general.
 * 
 * @author dklotz
 * @param <T>
 *            The type of the two objects.
 */
public class Pair<T> {

    /** The first object of the pair. */
    private final T first;
    /** The second object of the pair. */
    private final T second;

    /**
     * Creates a new pair.
     * 
     * @param first
     *            The first object.
     * @param second
     *            The second object.
     */
    public Pair(T first, T second) {
        this.first = first;
        this.second = second;
    }

    /**
     * Returns the first entry of the pair.
     * 
     * @return first entry
     */
    public T getFirst() {
        return first;
    }

    /**
     * Returns the second entry of the pair.
     * 
     * @return second entry
     */
    public T getSecond() {
        return second;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @SuppressWarnings("unchecked")
    @Override
    public boolean equals(Object obj) {
        // We can only be equal to other pairs.
        if (!(obj instanceof Pair)) {
            return false;
        }

        Pair<?> other = (Pair<?>) obj;

        // We are equal if both objects are equal.
        return getFirst().equals(other.getFirst())
                && getSecond().equals(other.getSecond());
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode() {
        // Return a combination of both objects hashcodes.
        // The factors are there so this is not symmetric (the order of first
        // and second matters).
        // CHECKSTYLE:OFF
        return 32 * getFirst().hashCode() + 42 * getSecond().hashCode();
        // CHECKSTYLE:ON
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        String firstString = null;
        if (getFirst() != null) {
            firstString = getFirst().toString();
        }

        String secondString = null;
        if (getSecond() != null) {
            secondString = getSecond().toString();
        }

        return "[" + firstString + ", " + secondString + "]";
    }

}
