/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.util.slam;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.event.MouseEvent;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.Map;

import javax.swing.JFrame;

import org.apache.commons.collections15.Transformer;
import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.util.Pair;
import de.unibi.airobots.bonsai.util.slam.gui.MapPanel;
import de.unibi.airobots.bonsai.util.slam.gui.MapPanel.WorldClickListener;
import de.unibi.airobots.btl.data.BinarySlamMap;
import de.unibi.airobots.btl.data.DynamicGridMap;
import de.unibi.airobots.btl.data.GlobalPlan;
import de.unibi.airobots.btl.data.NavigationGoalData;
import de.unibi.airobots.btl.data.NavigationGoalData.GoalType;
import de.unibi.airobots.btl.data.PositionData;
import edu.uci.ics.jung.algorithms.shortestpath.DijkstraDistance;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.UndirectedSparseMultigraph;

/**
 * Some utility methods for working with SLAM data.
 * 
 * @author dklotz
 */
public class SlamTools {

    private static final int DEFAULT_UPDATE_TIME = 500;
    private static final float SLAM_MIDDLE_VALUE = 0.5f;
    private static final float SLAM_TOLERANCE = 0.001f;
    private static final double WEIGHT_STRAIGHT = 1.0;
    private static final double WEIGHT_DIAGONAL = Math.sqrt(2.0);

    /** For logging output. */
    private static final Logger LOGGER = Logger.getLogger(SlamTools.class);

    /**
     * Converts the slam map contained in the {@link BinarySlamMapData} to a
     * displayable image.
     * 
     * @param slamMap
     *            The SLAM data container, must not be null.
     * @return An image of the occupancy grid in its original size (1 grid cell
     *         = 1 pixel), never null.
     */
    public static BufferedImage slamMapToImage(BinarySlamMap slamMap) {

        return gridMapToImage(slamMap.getDynamicGridMap());
    }

    public static BufferedImage gridMapToImage(DynamicGridMap map) {
        // Graphics setup, gray level image etc.
        BufferedImage image =
                new BufferedImage(map.getWidth(), map.getHeight(),
                        BufferedImage.TYPE_BYTE_GRAY);
        Graphics2D g2d = image.createGraphics();

        // Create one pixel for each point in the occupancy grid
        for (int x = 0; x < map.getWidth(); x++) {
            for (int y = 0; y < map.getHeight(); y++) {
                // Create a grey value from the occupancy value (0 = black =
                // occupied, 1 = white = unoccupied).
                float mapValue = map.getGridMap()[x][y];
                g2d.setColor(new Color(mapValue, mapValue, mapValue));

                // Create a 1 pixel rectangle and draw it (mirroring the y
                // coordinate, because of the occupancy grids layout, described
                // at the top).
                Rectangle2D pixelRect =
                        new Rectangle2D.Double(x, (map.getHeight() - 1) - y, 1,
                                1);
                g2d.draw(pixelRect);
            }
        }

        return image;
    }

    /**
     * This is just a convenience method for creating and showing a window
     * containing a {@link MapPanel} with one registered
     * {@link WorldClickListener} that sends goals of the clicked positions to
     * the robot.
     * 
     * @param slamSensor
     *            The sensor used to get the SLAM map.
     * @param posSensor
     *            position sensor to use for displaying the current position of
     *            the robot
     * @param planSensor
     *            sensor to use for displaying the current global plan
     * @param actuator
     *            The actuator used to make the robot move.
     * @return The (already visible) window, which will by default dispose on
     *         close. You don't necessarily need to do anything with it.
     */
    public static MapPanel showMapPanelWindow(
            Sensor<BinarySlamMap> slamSensor,
            Sensor<PositionData> posSensor, final NavigationActuator actuator,
            final Sensor<GlobalPlan> planSensor) {
        return showMapPanelWindow(slamSensor, posSensor, actuator, planSensor,
                true);
    }

    /**
     * This is just a convenience method for creating and showing a window
     * containing a {@link MapPanel} with one registered
     * {@link WorldClickListener} that sends goals of the clicked positions to
     * the robot.
     * 
     * @param slamSensor
     *            The sensor used to get the SLAM map.
     * @param posSensor
     *            position sensor to display the current robot position
     * @param actuator
     *            The actuator used to make the robot move.
     * @param planSensor
     *            sensor to display the current global plan
     * @param registerNavigator
     *            Should we register the default listener that navigates to the
     *            clicked position?
     * @return The (already visible) panel, which will by default dispose on
     *         close. You don't necessarily need to do anything with it.
     */
    public static MapPanel showMapPanelWindow(
            Sensor<BinarySlamMap> slamSensor,
            Sensor<PositionData> posSensor, final NavigationActuator actuator,
            final Sensor<GlobalPlan> planSensor, boolean registerNavigator) {

        // Create the map panel
        MapPanel mapPanel =
                new MapPanel(slamSensor, actuator, posSensor, planSensor,
                        DEFAULT_UPDATE_TIME);

        if (registerNavigator) {
            // Create and register the click listener
            MapPanel.WorldClickListener navigator =
                    new MapPanel.WorldClickListener() {

                        @Override
                        public void worldClicked(Pair<Double> worldCoordinates,
                                MouseEvent mouseEvent) {

                            NavigationGoalData goal = new NavigationGoalData();

                            goal.setX(worldCoordinates.getFirst());
                            goal.setY(worldCoordinates.getSecond());

                            if (mouseEvent.getButton() == MouseEvent.BUTTON2
                                    || mouseEvent.getButton() == MouseEvent.BUTTON3) {
                                goal.setType(GoalType.GLOBAL);
                            }

                            goal.setYawTolerance(Math.PI * 2);

                            try {
                                actuator.setGoal(goal);
                            } catch (IOException e) {
                                LOGGER.warn("Could not set target!", e);
                            }
                        }
                    };

            mapPanel.addWorldClickListener(navigator);
        }
        // Create and show a window
        JFrame testFrame = new JFrame("SLAM MapPanel");
        testFrame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        testFrame.getContentPane().setLayout(new BorderLayout());
        testFrame.getContentPane().add(mapPanel);
        testFrame.pack();
        testFrame.setVisible(true);

        return mapPanel;
    }

    /**
     * A simple node for the navigation graph.
     * 
     * @author lziegler
     */
    public static class Node {

        private static int nodeCount = 0;
        private int id;
        private int anchorX;
        private int anchorY;
        private boolean clear;

        public Node(int anchorX, int anchorY, boolean clear) {
            this.id = nodeCount++;
            this.anchorX = anchorX;
            this.anchorY = anchorY;
            this.clear = clear;
        }

        public boolean isclear() {
            return clear;
        }

        public int getId() {
            return id;
        }

        public int getAnchorX() {
            return anchorX;
        }

        public int getAnchorY() {
            return anchorY;
        }

        public String toString() {
            return "V" + id;
        }
    }

    /**
     * A simple edge for the navigation graph.
     * 
     * @author lziegler
     */
    public static class Edge {

        private double weight;
        private static int edgeCount = 0;
        private int id;

        public Edge(double weight) {
            this.id = edgeCount++;
            this.weight = weight;
        }

        public double getWeight() {
            return weight;
        }

        public int getId() {
            return id;
        }

        public String toString() {
            return "E" + id;
        }
    }

    /**
     * Calculates the nearest node to one map bin.
     * 
     * @param g
     *            The graph.
     * @param x
     *            The map bin's x coordinate.
     * @param y
     *            The map bin's y coordinate.
     * @return The corresponding node.
     */
    public static Node getNearestNode(Graph<Node, Edge> g, int x, int y) {

        double minDist = Double.MAX_VALUE;
        Node nearest = null;
        for (Node node : g.getVertices()) {
            double dx = x - node.getAnchorX();
            double dy = y - node.getAnchorY();
            double dist = Math.sqrt(Math.pow(dx, 2) + Math.pow(dy, 2));
            if (dist < minDist) {
                minDist = dist;
                nearest = node;
            }
        }
        return nearest;
    }

    /**
     * Calculates the shortest path distance to a target bin.
     * 
     * @param g
     *            The graph.
     * @param distanceMap
     *            The distance map.
     * @param targetX
     *            The target map bin's x coordinate.
     * @param targetY
     *            The target map bin's y coordinate.
     * @return the shortest path distance to the given bin.
     */
    public static double getDistance(Graph<Node, Edge> g,
            Map<Node, Number> distanceMap, int targetX, int targetY) {

        // TODO this can be implemented much faster!!!
        Node target = getNearestNode(g, targetX, targetY);
        if (distanceMap.containsKey(target)) {
            return distanceMap.get(target).doubleValue();
        } else {
            // not reachable
            return Double.MAX_VALUE;
        }
    }

    /**
     * Generates a map containing the shortest path distances to each node
     * relative to one reference bin.
     * 
     * @param g
     *            The graph.
     * @param sourceX
     *            The reference bin's x coordinate.
     * @param sourceY
     *            The reference bin's y coordinate.
     * @return The distance map.
     */
    public static Map<Node, Number> getDistanceMap(Graph<Node, Edge> g,
            int sourceX, int sourceY) {

        Transformer<Edge, Double> transformer =
                new Transformer<Edge, Double>() {

                    public Double transform(Edge link) {
                        return link.getWeight();
                    }
                };

        DijkstraDistance<Node, Edge> distanceAlgo =
                new DijkstraDistance<SlamTools.Node, SlamTools.Edge>(g,
                        transformer);

        Node source = getNearestNode(g, sourceX, sourceY);

        return distanceAlgo.getDistanceMap(source);
    }

    /**
     * Generates a graph corresponding to the given slam map, that can be used
     * for different graph algorithms.
     * 
     * @param slamMap
     *            The slam map.
     * @param numBinsPerNode
     *            The number of slam grid map bins, that should belong to one
     *            node on one axis. That means one node will correspond to
     *            numBinsPerNode^2 map bins.
     * @return The graph.
     */
    public static Graph<Node, Edge> generateNavigationGraph(
            DynamicGridMap slamMap, int numBinsPerNode) {

        UndirectedSparseMultigraph<Node, Edge> g =
                new UndirectedSparseMultigraph<Node, Edge>();
        int halfSize = numBinsPerNode / 2;
        Node[][] nodeMap = new Node[slamMap.getWidth()][slamMap.getHeight()];

        // traverse map
        for (int x = 0; x <= slamMap.getWidth() - numBinsPerNode; x +=
                numBinsPerNode) {
            for (int y = 0; y <= slamMap.getHeight() - numBinsPerNode; y +=
                    numBinsPerNode) {

                // check if area is clear
                boolean clear = isAreaClear(slamMap, x, y, numBinsPerNode);

                // add node
                Node n = new Node(x + halfSize / 2, y + halfSize / 2, clear);
                g.addVertex(n);
                nodeMap[x][y] = n;

                // edge to left node
                if (x - numBinsPerNode >= 0) {
                    Node otherNode = nodeMap[x - numBinsPerNode][y];
                    if (n.isclear() && otherNode.isclear()) {
                        g.addEdge(new Edge(WEIGHT_STRAIGHT), n, otherNode);
                    }
                }

                // edge to upper node
                if (y - numBinsPerNode >= 0) {
                    Node otherNode = nodeMap[x][y - numBinsPerNode];
                    if (n.isclear() && otherNode.isclear()) {
                        g.addEdge(new Edge(WEIGHT_STRAIGHT), n, otherNode);
                    }
                }

                // edge to upper left node
                if (y - numBinsPerNode >= 0 && x - numBinsPerNode >= 0) {
                    Node otherNode =
                            nodeMap[x - numBinsPerNode][y - numBinsPerNode];
                    if (n.isclear() && otherNode.isclear()) {
                        g.addEdge(new Edge(WEIGHT_DIAGONAL), n, otherNode);
                    }
                }

                // edge to beneath left node
                if (y + numBinsPerNode < slamMap.getHeight()
                        && x - numBinsPerNode >= 0) {
                    Node otherNode =
                            nodeMap[x - numBinsPerNode][y + numBinsPerNode];
                    if (n.isclear() && otherNode.isclear()) {
                        g.addEdge(new Edge(WEIGHT_DIAGONAL), n, otherNode);
                    }
                }
            }
        }

        return g;
    }

    private static boolean isAreaClear(DynamicGridMap slamMap, int topLeftX,
            int topLeftY, int areaSize) {

        boolean foundObstacle = false;
        boolean foundUnknown = false;
        boolean foundClear = false;
        for (int x = topLeftX; x < topLeftX + areaSize; x++) {
            if (x >= slamMap.getWidth()) {
                break;
            }
            for (int y = topLeftY; y < topLeftY + areaSize; y++) {
                if (y >= slamMap.getHeight()) {
                    break;
                }
                if (slamMap.getValue(x, y) < SLAM_MIDDLE_VALUE - SLAM_TOLERANCE) {
                    foundObstacle = true;
                } else if (slamMap.getValue(x, y) < SLAM_MIDDLE_VALUE
                        + SLAM_TOLERANCE) {
                    foundUnknown = true;
                } else {
                    foundClear = true;
                }
            }
        }
        if (foundClear && !foundObstacle) {
            return true;
        } else {
            return false;
        }
    }

    private SlamTools() {
        // Just to prevent instantiation.
    }
}
